// ObjectNode.cpp
//------------------------------------------------------------------------------

// Includes
//------------------------------------------------------------------------------
#include "ObjectNode.h"

#include <Tools/FBuild/FBuildCore/FBuild.h>
#include <Tools/FBuild/FBuildCore/Flog.h>
#include <Tools/FBuild/FBuildCore/Graph/NodeGraph.h>
#include <Tools/FBuild/FBuildCore/Helpers/CIncludeParser.h>
#include <Tools/FBuild/FBuildCore/Helpers/Compressor.h>
#include <Tools/FBuild/FBuildCore/WorkerPool/Job.h>
#include <Tools/FBuild/FBuildCore/WorkerPool/JobQueue.h>
#include <Tools/FBuild/FBuildCore/WorkerPool/WorkerThread.h>

#include <Core/FileIO/FileIO.h>
#include <Core/FileIO/FileStream.h>
#include <Core/Math/CRC32.h>
#include <Core/Process/Process.h>
#include <Core/Process/Thread.h>
#include <Core/Tracing/Tracing.h>
#include <Core/Strings/AStackString.h>

#include <string.h>
#include <windows.h> // for GetLastError


// CONSTRUCTOR
//------------------------------------------------------------------------------
ObjectNode::ObjectNode( const AString & objectName,
					    Node * inputNode,
					    Node * compilerNode,
						const AString & compilerArgs,
						Node * precompiledHeader,
						uint32_t flags,
						const Array< Node * > & compilerForceUsing )
: FileNode( objectName, Node::FLAG_NONE )
, m_Includes( 0, true )
, m_Flags( flags )
, m_CompilerArgs( compilerArgs )
, m_CompilerForceUsing( compilerForceUsing )
{
	m_StaticDependencies.SetCapacity( 3 );

	ASSERT( compilerNode );
	m_StaticDependencies.Append( compilerNode );

	ASSERT( inputNode );
	m_StaticDependencies.Append( inputNode );

	if ( precompiledHeader )
	{
		m_StaticDependencies.Append( precompiledHeader );
	}

	m_StaticDependencies.Append( compilerForceUsing );

	m_Type = OBJECT_NODE;
	m_LastBuildTimeMs = 5000; // higher default than a file node
}

// DESTRUCTOR
//------------------------------------------------------------------------------
ObjectNode::~ObjectNode()
{
}

// DoBuild
//------------------------------------------------------------------------------
/*virtual*/ Node::BuildResult ObjectNode::DoBuild( Job * job )
{
	// delete previous file
	if ( FileIO::FileExists( GetName().Get() ) )
	{
		if ( FileIO::FileDelete( GetName().Get() ) == false )
		{
			FLOG_ERROR( "Failed to delete file before build '%s'", GetName().Get() );
			return NODE_RESULT_FAILED;
		}
	}
	if ( GetFlag( FLAG_MSVC ) && GetFlag( FLAG_CREATING_PCH ) )
	{
		AStackString<> pchObj( GetName() );
		pchObj += GetObjExtension();
		if ( FileIO::FileExists( pchObj.Get() ) )
		{
			if ( FileIO::FileDelete( pchObj.Get() ) == false )
			{
				FLOG_ERROR( "Failed to delete file before build '%s'", GetName().Get() );
				return NODE_RESULT_FAILED;
			}
		}
	}

	bool useCache = GetFlag( FLAG_CAN_BE_CACHED ) &&
					( FBuild::Get().GetOptions().m_UseCacheRead ||
					  FBuild::Get().GetOptions().m_UseCacheWrite );
	bool useDist = GetFlag( FLAG_CAN_BE_DISTRIBUTED ) && FBuild::Get().GetOptions().m_AllowDistributed;
	bool usePreProcessor = ( useCache || useDist || GetFlag( FLAG_GCC ) || GetFlag( FLAG_SNC ) );

	if ( usePreProcessor )
	{
		return DoBuildWithPreProcessor( job );
	}

	if ( GetFlag( FLAG_MSVC ) )
	{
		return DoBuildMSCL_NoCache();
	}

	return DoBuildOther();
}

// DoBuild_Remote
//------------------------------------------------------------------------------
/*virtual*/ Node::BuildResult ObjectNode::DoBuild2( Job * job )
{
	return DoBuildWithPreProcessor2( job );
}

// Finalize
//------------------------------------------------------------------------------
/*virtual*/ bool ObjectNode::Finalize()
{
	ASSERT( Thread::IsMainThread() );

	NodeGraph & ng = FBuild::Get().GetDependencyGraph();

	// convert includes to nodes
	m_DynamicDependencies.Clear();
	m_DynamicDependencies.SetCapacity( m_Includes.GetSize() );
	for ( Array< AString >::ConstIter it = m_Includes.Begin();
			it != m_Includes.End();
			it++ )
	{
		Node * fn = FBuild::Get().GetDependencyGraph().FindNode( *it );
		if ( fn == nullptr )
		{
			fn = ng.CreateFileNode( *it );
		}
		else if ( fn->IsAFile() == false )
		{
			FLOG_ERROR( "'%s' is not a FileNode (type: %s)", fn->GetName().Get(), fn->GetTypeName() );
			return false;
		}
		m_DynamicDependencies.Append( fn );
	}

	return true;
}

// DoBuildMSCL_NoCache
//------------------------------------------------------------------------------
/*virtual*/ Node::BuildResult ObjectNode::DoBuildMSCL_NoCache()
{
	// Format compiler args string
	AStackString< 8 * KILOBYTE > fullArgs;
	BuildFullArgs( fullArgs, false );
	fullArgs += " /showIncludes"; // we'll extract dependency information from this

	FileNode * compiler = GetCompiler()->CastTo< FileNode >();

	EmitCompilationMessage( fullArgs );

//  TODO:C Use this
//	if ( BuildFinalOutput( fullArgs ) == false )
//	{
//		return NODE_RESULT_FAILED; // BuildFinalOutput will have emitted error
//	}

	// spawn the process
	Process p;
	if ( false == p.Spawn( compiler->GetName().Get(), 
						   fullArgs.Get(),
						   nullptr, // workingDir - use the exe launch dir
						   FBuild::Get().GetEnvironmentString() ) )
	{
		FLOG_ERROR( "Failed to spawn process to build '%s'", GetName().Get() );
		return NODE_RESULT_FAILED;
	}

	// capture all of the stdout and stderr
	AutoPtr< char > memOut;
	AutoPtr< char > memErr;
	uint32_t memOutSize = 0;
	uint32_t memErrSize = 0;
	p.ReadAllData( memOut, &memOutSize, memErr, &memErrSize );

	// Get result
	ASSERT( !p.IsRunning() );
	int result = p.WaitForExit();

	if ( memErr.Get() )
	{ 
		DumpOutput( memErr.Get(), memErrSize );
	}

	if ( result != 0 )
	{
		DumpOutput( memOut.Get(), memOutSize );
		FLOG_ERROR( "Failed to build Object (error %i) '%s'", result, GetName().Get() );
		return NODE_RESULT_FAILED;
	}

	// compiled ok, try to extract includes
	if ( ProcessIncludesMSCL( memOut.Get(), memOutSize, false) == false )
	{
		return NODE_RESULT_FAILED; // ProcessIncludesMSCL will have emitted an error
	}

	// record new file time
	m_Stamp = FileIO::GetFileLastWriteTime( m_Name );

	return NODE_RESULT_OK;
}

// DoBuildWithPreProcessor
//------------------------------------------------------------------------------
Node::BuildResult ObjectNode::DoBuildWithPreProcessor( Job * job )
{
	AStackString< 8 * KILOBYTE > fullArgs;
	BuildFullArgs( fullArgs, true );

	if ( BuildPreprocessedOutput( fullArgs, job ) == false )
	{
		return NODE_RESULT_FAILED; // BuildPreprocessedOutput will have emitted an error
	}

	// preprocessed ok, try to extract includes
	if ( ProcessIncludesWithPreProcessor( job ) == false )
	{
		return NODE_RESULT_FAILED; // ProcessIncludesWithPreProcessor will have emitted an error
	}

	// calculate the cache entry lookup
	AStackString<> cacheFile;
	if ( FBuild::Get().GetOptions().m_UseCacheRead ||
		 FBuild::Get().GetOptions().m_UseCacheWrite )
	{
		GetCacheName( job->GetData(), job->GetDataSize(), cacheFile );
	}

	// try to get from cache
	if ( RetrieveFromCache( cacheFile ) )
	{
		return NODE_RESULT_OK_CACHE;
	}

	// can we do the rest of the work remotely?
	if ( GetFlag( FLAG_CAN_BE_DISTRIBUTED ) &&
		 FBuild::Get().GetOptions().m_AllowDistributed &&
		 JobQueue::Get().GetDistributableJobsMemUsage() < ( 512 * MEGABYTE ) )
	{
		// compress job data
		Compressor c;
		c.Compress( job->GetData(), job->GetDataSize() );
		size_t compressedSize = c.GetResultSize();
		job->OwnData( c.ReleaseResult(), compressedSize, true );

		// yes... re-queue for secondary build
		return NODE_RESULT_NEED_SECOND_BUILD_PASS;
	}

	// can't do the work remotely, so do it right now
	Node::BuildResult result = DoBuildWithPreProcessor2( job );
	if ( result != Node::NODE_RESULT_OK )
	{
		return result;
	}

	// record new file time
	m_Stamp = FileIO::GetFileLastWriteTime( m_Name );

	if ( m_Stamp && FBuild::Get().GetOptions().m_UseCacheWrite )
	{
		WriteToCache( cacheFile );
	}

	return Node::NODE_RESULT_OK;
}

// DoBuildWithPreProcessor2
//------------------------------------------------------------------------------
Node::BuildResult ObjectNode::DoBuildWithPreProcessor2( Job * job )
{
	AStackString< 8 * KILOBYTE > fullArgs;
	BuildFullArgs( fullArgs, false );

	bool usePreProcessedOutput = true;

	if ( GetFlag( FLAG_MSVC ) && GetFlag( FLAG_USING_PCH ) && job->IsLocal() )
	{
		// for MSVC, if using precompiled headers, we don't want to compile from the preprocessed
		// output, as (assuming it's well tuned) it's quicker to build using the precompiled header
		// (but we can only do this if building locally, as remotely we don't have access to the
		// pch)
		usePreProcessedOutput = false;
	}
	
	if ( usePreProcessedOutput )
	{
		AStackString<> tmpFileName;
		if ( WriteTmpFile( job, tmpFileName ) == false )
		{
			return NODE_RESULT_FAILED; // WriteTmpFile will have emitted an error
		}

		// use preprocesed output path as input
		Node * sourceFile = GetSourceFile();
		fullArgs.Replace( sourceFile->GetName().Get(), tmpFileName.Get() );

		// remove pch related flags ('Yu' and 'Fp')
		if ( GetFlag( FLAG_USING_PCH ) )
		{
			Array< AString > args;
			fullArgs.Tokenize( args );
			fullArgs.Clear();
			auto iter = args.Begin();
			const auto end = args.End();
			for ( ; iter != end; ++iter )
			{
				if ( iter->BeginsWith( "/Yu" ) ||
					 iter->BeginsWith( "/Fp" ) )
				{
					continue;
				}
				if ( iter != args.Begin() )
				{
					fullArgs += ' ';
				}
				fullArgs += *iter;
			}
		}

	}

	if ( BuildFinalOutput( job, fullArgs ) == false )
	{
		return NODE_RESULT_FAILED; // BuildFinalOutput will have emitted error
	}
	return NODE_RESULT_OK;
}

// DoBuildOther
//------------------------------------------------------------------------------
/*virtual*/ Node::BuildResult ObjectNode::DoBuildOther()
{
	// Format compiler args string
	AStackString< 8 * KILOBYTE > fullArgs;
	BuildFullArgs( fullArgs, false );

	EmitCompilationMessage( fullArgs );

	FileNode * compiler = GetCompiler()->CastTo< FileNode >();

	// spawn the process
	Process p;
	if ( false == p.Spawn( compiler->GetName().Get(), 
						   fullArgs.Get(),
						   nullptr, // use exe launch dir as the working dir 
						   FBuild::Get().GetEnvironmentString() ) )
	{
		FLOG_ERROR( "Failed to spawn process to build '%s'", GetName().Get() );
		return NODE_RESULT_FAILED;
	}

	// capture all of the stdout and stderr
	AutoPtr< char > memOut;
	AutoPtr< char > memErr;
	uint32_t memOutSize = 0;
	uint32_t memErrSize = 0;
	p.ReadAllData( memOut, &memOutSize, memErr, &memErrSize );

	// Get result
	ASSERT( !p.IsRunning() );
	int result = p.WaitForExit();
	bool ok = ( result == 0 );

	if ( memErr.Get() )
	{ 
		DumpOutput( memErr.Get(), memErrSize );
	}

	if ( !ok )
	{
		// something went wrong, print details
		DumpOutput( memOut.Get(), memOutSize );
		FLOG_ERROR( "Failed to build Object (error %i) '%s'", result, GetName().Get() );
		return NODE_RESULT_FAILED;
	}

	// record new file time
	m_Stamp = FileIO::GetFileLastWriteTime( m_Name );

	return NODE_RESULT_OK;
}

// ProcessIncludesMSCL
//------------------------------------------------------------------------------
bool ObjectNode::ProcessIncludesMSCL( const char * output, uint32_t outputSize, bool preprocessed )
{
	Timer t;

	{
		ASSERT( output && outputSize );

		CIncludeParser parser;
		bool result = preprocessed ? parser.ParseMSCL_Preprocessed( output, outputSize )
								   : parser.ParseMSCL_Output( output, outputSize );
		if ( result == false )
		{
			FLOG_ERROR( "Failed to process includes for '%s'", GetName().Get() );
			return false;
		}

		// record that we have a list of includes
		// (we need a flag because we can't use the array size
		// as a determinator, because the file might not include anything)
		m_Includes.Clear();
		parser.SwapIncludes( m_Includes );
	}

	FLOG_INFO( "Process Includes:\n - File: %s\n - Time: %u ms\n - Num : %u", m_Name.Get(), uint32_t( t.GetElapsedMS() ), uint32_t( m_Includes.GetSize() ) );

	return true;
}

// ProcessIncludesWithPreProcessor
//------------------------------------------------------------------------------
bool ObjectNode::ProcessIncludesWithPreProcessor( Job * job )
{
	Timer t;

	{
		const char  * output = (char *)job->GetData();
		const size_t outputSize = job->GetDataSize();

		ASSERT( output && outputSize );

		CIncludeParser parser;
		bool result = GetFlag( FLAG_MSVC ) ? parser.ParseMSCL_Preprocessed( output, outputSize )
										   : parser.ParseGCC_Preprocessed( output, outputSize );
		if ( result == false )
		{
			FLOG_ERROR( "Failed to process includes for '%s'", GetName().Get() );
			return false;
		}

		// record that we have a list of includes
		// (we need a flag because we can't use the array size
		// as a determinator, because the file might not include anything)
		m_Includes.Clear();
		parser.SwapIncludes( m_Includes );
	}

	FLOG_INFO( "Process Includes:\n - File: %s\n - Time: %u ms\n - Num : %u", m_Name.Get(), uint32_t( t.GetElapsedMS() ), uint32_t( m_Includes.GetSize() ) );

	return true;
}

// Load
//------------------------------------------------------------------------------
/*static*/ Node * ObjectNode::Load( IOStream & stream, bool remote )
{
	NODE_LOAD( AStackString<>,	name );
	NODE_LOAD_DEPS( 3,			staticDeps );
	NODE_LOAD_DEPS( 0,			dynamicDeps );
	NODE_LOAD( uint32_t,		flags );
	NODE_LOAD( AStackString<>,	compilerArgs );
	NODE_LOAD( AStackString<>,	objExtensionOverride );
	NODE_LOAD_DEPS( 0,			compilerForceUsing )

	// we are making inferences from the size of the staticDeps
	// ensure we catch if those asumptions break
	size_t numStaticDepsExcludingForceUsing = staticDeps.GetSize() - compilerForceUsing.GetSize();
	ASSERT( ( numStaticDepsExcludingForceUsing == 2 ) ||
 			( numStaticDepsExcludingForceUsing == 3 ) );

	ASSERT( staticDeps.GetSize() >= 2 );
	Node * compiler = staticDeps[ 0 ];
	Node * staticDepNode = staticDeps[ 1 ];

	Node * precompiledHeader = nullptr;
	if ( numStaticDepsExcludingForceUsing == 3 )
	{
		precompiledHeader = staticDeps[ 2 ];
	}

	Node * on = nullptr;
	if ( remote )
	{
		AStackString<> dummyName( "$$dummyName$$" );
		on = new ObjectNode( dummyName, staticDepNode, compiler, compilerArgs, precompiledHeader, flags, compilerForceUsing );
	}
	else
	{
		NodeGraph & ng = FBuild::Get().GetDependencyGraph();
		on = ng.CreateObjectNode( name, staticDepNode, compiler, compilerArgs, precompiledHeader, flags, compilerForceUsing );
	}

	ObjectNode * objNode = on->CastTo< ObjectNode >();
	objNode->m_DynamicDependencies.Swap( dynamicDeps );
	objNode->m_ObjExtensionOverride = objExtensionOverride;

	return objNode;
}

// Save
//------------------------------------------------------------------------------
/*virtual*/ bool ObjectNode::Save( IOStream & stream, bool remote ) const
{
	NODE_SAVE( m_Name );
	NODE_SAVE_DEPS( m_StaticDependencies );
	NODE_SAVE_DEPS( m_DynamicDependencies );
	NODE_SAVE( m_Flags );
	NODE_SAVE( m_CompilerArgs );
	NODE_SAVE( m_ObjExtensionOverride );
	NODE_SAVE_DEPS( m_CompilerForceUsing )

	return true;
}

// GetObjExtension
//------------------------------------------------------------------------------
const char * ObjectNode::GetObjExtension() const
{
	if ( m_ObjExtensionOverride.IsEmpty() )
	{
		return ".obj";
	}
	return m_ObjExtensionOverride.Get();
}

// DumpOutput
//------------------------------------------------------------------------------
/*static*/ void ObjectNode::DumpOutput( const char * data, uint32_t dataSize )
{
	if ( ( data != nullptr ) && ( dataSize > 0 ) )
	{
		Array< AString > exclusions( 2, false );
		exclusions.Append( AString( "Note: including file:" ) );
		exclusions.Append( AString( "#line" ) );
		Node::DumpOutput( data, dataSize, &exclusions );
	}
}

// GetCacheName
//------------------------------------------------------------------------------
void ObjectNode::GetCacheName( const void * memOut, size_t memOutSize,
										  AString & cacheName ) const
{
	Timer t;

	// hash the pre-processed intput data
	uint32_t a = CRC32::Calc( memOut, memOutSize );

	// hash the build "environment"
	uint32_t b = CRC32::Start();
	b = CRC32::Update( b, m_CompilerArgs.Get(), m_CompilerArgs.GetLength() );
	// TODO:B Exclude preprocessor control defines (the preprocessed input has considered those already)
	// TODO:B - Include compiler version/compiler hash
	b = CRC32::Stop( b );

	FBuild::Get().GetCacheFileName( a, b, cacheName );

	FLOG_INFO( "Cache hash: %u ms - %u kb '%s'\n", 
				uint32_t( t.GetElapsedMS() ), 
				uint32_t( memOutSize / KILOBYTE ), 
				cacheName.Get() );
}

// RetrieveFromCache
//------------------------------------------------------------------------------
bool ObjectNode::RetrieveFromCache( const AString & cacheFileName )
{
	if ( FBuild::Get().GetOptions().m_UseCacheRead == false )
	{
		return false;
	}

	Timer t;
	FileStream cacheFile;
	if ( cacheFile.Open( cacheFileName.Get(), FileStream::READ_ONLY ) )
	{
		const size_t cacheFileSize = (size_t)cacheFile.GetFileSize();
		AutoPtr< char > mem( (char *)::Alloc( cacheFileSize ) );
		if ( cacheFile.Read( mem.Get(), cacheFileSize ) == cacheFileSize )
		{
			// we expect a cache file header indicating compression status
			const uint32_t headerVal = *( (uint32_t *)mem.Get() );
			ASSERT( ( headerVal == 'OBJ0' ) || ( headerVal == 'OBJ1' ) );

			FileStream objFile;
			if ( !objFile.Open( m_Name.Get(), FileStream::WRITE_ONLY ) )
			{
				FLOG_ERROR( "Failed to open local file during cache retrieval '%s'", m_Name.Get() );
				return false;
			}

			// header contains compressed state
			const bool compressed = ( headerVal == 'OBJ1' );
			const void * data = ( mem.Get() + sizeof( uint32_t ) );
			size_t dataSize = cacheFileSize - sizeof( uint32_t );
			Compressor c;
			if ( compressed )
			{
				c.Decompress( data );
				data = c.GetResult();
				dataSize = c.GetResultSize();
			}

			if ( objFile.Write( data, dataSize ) != dataSize )
			{
				FLOG_ERROR( "Failed to write to local file during cache retrieval '%s'", m_Name.Get() );
				return false;
			}

			// Get current "system time" and convert to "file time"
			SYSTEMTIME st;
			FILETIME ft;
			GetSystemTime( &st );
			if ( FALSE == SystemTimeToFileTime( &st, &ft ) )
			{
				FLOG_ERROR( "Failed to convert file time after cache hit '%s' (%u)", m_Name.Get(), GetLastError() );
				return false;
			}
			uint64_t fileTimeNow = ( (uint64_t)ft.dwLowDateTime | ( (uint64_t)ft.dwHighDateTime << 32 ) );
	
			// set the time on the local file
			if ( objFile.SetLastWriteTime( fileTimeNow ) == false )
			{
				FLOG_ERROR( "Failed to set timestamp on file after cache hit '%s' (%u)", m_Name.Get(), GetLastError() );
				return false;
			}

			objFile.Close();

			FileIO::WorkAroundForWindowsFilePermissionProblem( m_Name );

			// the file time we set and local file system might have different 
			// granularity for timekeeping, so we need to update with the actual time written
			m_Stamp = FileIO::GetFileLastWriteTime( m_Name );

			FLOG_INFO( "Cache hit: %u ms '%s'\n", uint32_t( t.GetElapsedMS() ), cacheFileName.Get() );
			SetStatFlag( Node::STATS_CACHE_HIT );
			return true;
		}
	}

	FLOG_INFO( "Cache miss: %u ms '%s'\n", uint32_t( t.GetElapsedMS() ), cacheFileName.Get() );
	SetStatFlag( Node::STATS_CACHE_MISS );
	return false;
}

// WriteToCache
//------------------------------------------------------------------------------
void ObjectNode::WriteToCache( const AString & cacheFileName )
{
	ASSERT( !cacheFileName.IsEmpty() );

	Timer t;

	// make sure the cache output path exists
	AStackString<> cachePath( cacheFileName.Get(), cacheFileName.FindLast( '\\' ) );
	if ( !FileIO::EnsurePathExists( cachePath ) )
	{
		FLOG_INFO( "Cache store fail: %u ms '%s'\n", uint32_t( t.GetElapsedMS() ), cacheFileName.Get() );
		return;
	}

	// open the compiled object which we want to store to the cache
	FileStream objFile;
	if ( !objFile.Open( m_Name.Get(), FileStream::READ_ONLY ) )
	{
		FLOG_INFO( "Cache store fail: %u ms '%s'\n", uint32_t( t.GetElapsedMS() ), cacheFileName.Get() );
		return;
	}

	// open output cache (tmp) file
	AStackString<> cacheFileTmpName( cacheFileName );
	cacheFileTmpName += ".tmp";
	FileStream cacheTmpFile;
	if( !cacheTmpFile.Open( cacheFileTmpName.Get(), FileStream::WRITE_ONLY ) )
	{
		FLOG_INFO( "Cache store fail: %u ms '%s'\n", uint32_t( t.GetElapsedMS() ), cacheFileName.Get() );
		return;
	}

	// read obj all into memory
	const size_t objFileSize = (size_t)objFile.GetFileSize();
	AutoPtr< char > mem( (char *)::Alloc( objFileSize ) );
	if ( objFile.Read( mem.Get(), objFileSize ) != objFileSize )
	{
		FLOG_INFO( "Cache store fail: %u ms '%s'\n", uint32_t( t.GetElapsedMS() ), cacheFileName.Get() );
		return;
	}

	// try to compress
	Compressor c;
	const bool compressed = c.Compress( mem.Get(), objFileSize );
	const void * data = compressed ? c.GetResult() : mem.Get();
	const size_t dataSize = compressed ? c.GetResultSize() : objFileSize;

	// cache file contains a header to indicate compression
	const uint32_t headerVal = compressed ? 'OBJ1' : 'OBJ0';

	// write header and data
	bool cacheTmpWriteOk = cacheTmpFile.Write( headerVal ) && ( cacheTmpFile.Write( data, dataSize ) == dataSize );

	cacheTmpFile.Close();

	if ( !cacheTmpWriteOk )
	{
		// failed to write to cache tmp file
		FileIO::FileDelete( cacheFileTmpName.Get() ); // try to cleanup failure
		FLOG_INFO( "Cache store fail: %u ms '%s'\n", uint32_t( t.GetElapsedMS() ), cacheFileName.Get() );
		return;
	}

	// rename tmp file to real file
	if ( FileIO::FileMove( cacheFileTmpName, cacheFileName ) == false )
	{
		// try to delete (possibly) existing file
		FileIO::FileDelete( cacheFileName.Get() );

		// try rename again
		if ( FileIO::FileMove( cacheFileTmpName, cacheFileName ) == false )
		{
			// problem renaming file
			FileIO::FileDelete( cacheFileTmpName.Get() ); // try to cleanup tmp file
			FLOG_INFO( "Cache store fail: %u ms '%s'\n", uint32_t( t.GetElapsedMS() ), cacheFileName.Get() );
			return;
		}
	}

	// cache store complete
	FLOG_INFO( "Cache store: %u ms '%s'\n", uint32_t( t.GetElapsedMS() ), cacheFileName.Get() );
	SetStatFlag( Node::STATS_CACHE_STORE );
}

// EmitCompilationMessage
//------------------------------------------------------------------------------
void ObjectNode::EmitCompilationMessage( const AString & fullArgs ) const
{
	// print basic or detailed output, depending on options
	// we combine everything into one string to ensure it is contiguous in
	// the output
	AStackString<> output;
	output += "Obj: ";
	output += GetName();
	output += '\n';
	if ( FLog::ShowInfo() )
	{
		output += GetCompiler()->GetName();
		output += ' ';
		output += fullArgs;
		output += '\n';
	}
	FLOG_BUILD( "%s", output.Get() );
}

// BuildFullArgs
//------------------------------------------------------------------------------
void ObjectNode::BuildFullArgs( AString & fullArgs, bool preProcess ) const
{
	// Format compiler args string
	fullArgs = m_CompilerArgs;
	if ( preProcess )
	{
		if ( GetFlag( FLAG_MSVC ) )
		{
			fullArgs += " /E"; // run pre-processor only
		}
		else
		{
			ASSERT( GetFlag( FLAG_GCC ) || GetFlag( FLAG_SNC ) );
			fullArgs += " -E"; // run pre-processor only

			// for GCC/SNC we want to capture the preprocessing via stdout
			// so we must remove the output command
			fullArgs.Replace( "-o%2", "" );
		}
	}

	Array< AString > tokens( 1024, true );
	fullArgs.Tokenize( tokens );
	fullArgs.Clear();

	auto const end = tokens.End();
	for ( auto it = tokens.Begin(); it!=end; ++it )
	{
		const AString & token = *it;

		// %1 -> InputFile
		const char * found = token.Find( "%1" );
		if ( found )
		{
			fullArgs += AStackString<>( token.Get(), found );
			fullArgs += GetSourceFile()->GetName();
			fullArgs += AStackString<>( found + 2, token.GetEnd() );
			fullArgs += ' ';
			continue;
		}

		// %2 -> OutputFile
		found = token.Find( "%2" );
		if ( found )
		{
			fullArgs += AStackString<>( token.Get(), found );
			fullArgs += m_Name;
			fullArgs += AStackString<>( found + 2, token.GetEnd() );
			fullArgs += ' ';
			continue;
		}

		// %3 -> PrecompiledHeader Obj
		if ( GetFlag( FLAG_MSVC ) )
		{
			found = token.Find( "%3" );
			if ( found )
			{
				// handle /Option:%3 -> /Option:A
				fullArgs += AStackString<>( token.Get(), found );
				fullArgs += m_Name;
				fullArgs += GetObjExtension(); // convert 'PrecompiledHeader.pch' to 'PrecompiledHeader.pch.obj'
				fullArgs += AStackString<>( found + 2, token.GetEnd() );
				fullArgs += ' ';
				continue;
			}
		}

		// %4 -> CompilerForceUsing list
		if ( GetFlag( FLAG_MSVC ) )
		{
			found = token.Find( "%4" );
			if ( found )
			{
				AStackString<> pre( token.Get(), found );
				AStackString<> post( found + 2, token.GetEnd() );
				ExpandTokenList( m_CompilerForceUsing, fullArgs, pre, post );
				fullArgs += ' ';
				continue;
			}
		}

		// untouched token
		fullArgs += token;
		fullArgs += ' ';
	}
}

// ExpandTokenList
//------------------------------------------------------------------------------
void ObjectNode::ExpandTokenList( const Array< Node * > & nodes, AString & fullArgs, const AString & pre, const AString & post ) const
{
	const auto end = nodes.End();
	for ( auto it = nodes.Begin(); it != end; ++it )
	{
		Node * n = *it;

		fullArgs += pre;
		fullArgs += n->GetName();
		fullArgs += post;
		fullArgs += ' ';
	}
}

// BuildPreprocessedOutput
//------------------------------------------------------------------------------
bool ObjectNode::BuildPreprocessedOutput( const AString & fullArgs, Job * job ) const
{
	FileNode * compiler = GetCompiler()->CastTo< FileNode >();

	EmitCompilationMessage( fullArgs );

	// spawn the process
	Process p;
	if ( false == p.Spawn( compiler->GetName().Get(), 
						   fullArgs.Get(),
						   nullptr, // use exe launch dir as the working dir
						   FBuild::Get().GetEnvironmentString() ) )
	{
		FLOG_ERROR( "Failed to spawn process to build '%s'", GetName().Get() );
		return false;
	}

	// capture all of the stdout and stderr
	AutoPtr< char > memErr, memOut;
	uint32_t memErrSize = 0;
	uint32_t memOutSize = 0;
	p.ReadAllData( memOut, &memOutSize, memErr, &memErrSize );

	// Get result
	ASSERT( !p.IsRunning() );
	int result = p.WaitForExit();

	if ( result != 0 )
	{
		DumpOutput( memErr.Get(), memErrSize );
		FLOG_ERROR( "Failed to build Object (error %i) '%s'", result, GetName().Get() );
		return false;
	}

	// take a copy of the output because ReadAllData uses huge buffers to avoid re-sizing
	char * memCopy = (char *)::Alloc( memOutSize );
	memcpy( memCopy, memOut.Get(), memOutSize );

	job->OwnData( memCopy, memOutSize );

	return true;
}

// WriteTmpFile
//------------------------------------------------------------------------------
bool ObjectNode::WriteTmpFile( Job * job, AString & tmpFileName ) const
{
	ASSERT( job->GetData() && job->GetDataSize() );

	Node * sourceFile = GetSourceFile();

	FileStream tmpFile;
	const char * tmpFileExt = sourceFile->GetName().FindLast( '.' ) + 1;

	if ( GetFlag( FLAG_GCC ) || GetFlag( FLAG_SNC ) )
	{
		// GCC requires preprocessed output to be named a certain way
		// C code (.c) -> .i (NOTE: lower case only)
		// C++ code (.C, .cc, .cp, .cpp, .cxx) -> .ii
		tmpFileExt = ( strcmp( tmpFileExt, "c" ) == 0 ) ? "i" : "ii";
	}

	void const * dataToWrite = job->GetData();
	size_t dataToWriteSize = job->GetDataSize();

	// handle compressed data
	Compressor c; // scoped here so we can access decompression buffer
	if ( job->IsDataCompressed() )
	{
		c.Decompress( dataToWrite );
		dataToWrite = c.GetResult();
		dataToWriteSize = c.GetResultSize();
	}

	WorkerThread::CreateTempFilePath( tmpFileExt, tmpFileName );
	if ( WorkerThread::CreateTempFile( tmpFileName, tmpFile ) == false ) 
	{
		int error = GetLastError();
		FLOG_ERROR( "Failed to create temp file '%s' to build '%s' (error %u)", tmpFileName.Get(), GetName().Get(), error );
		return NODE_RESULT_FAILED;
	}
	if ( tmpFile.Write( dataToWrite, dataToWriteSize ) == false )
	{
		int error = GetLastError();
		FLOG_ERROR( "Failed to write to temp file '%s' to build '%s' (error %u)", tmpFileName.Get(), GetName().Get(), error );
		return NODE_RESULT_FAILED;
	}
	tmpFile.Close();

	FileIO::WorkAroundForWindowsFilePermissionProblem( tmpFileName );

	return true;
}

// BuildFinalOutput
//------------------------------------------------------------------------------
bool ObjectNode::BuildFinalOutput( Job * job, const AString & fullArgs ) const
{
	Node * compiler = GetCompiler();

	// spawn the process
	Process p;
	if ( false == p.Spawn( compiler->GetName().Get(), 
						   fullArgs.Get(),
						   nullptr, // workingDir - use the exe launch dir
						   FBuild::IsValid() ? FBuild::Get().GetEnvironmentString() : nullptr ) )
	{
		FLOG_ERROR( "Failed to spawn process to build '%s'", GetName().Get() );
		return false;
	}

	// capture all of the stdout and stderr
	AutoPtr< char > memErr, memOut;
	uint32_t memErrSize = 0;
	uint32_t memOutSize = 0;
	p.ReadAllData( memOut, &memOutSize, memErr, &memErrSize );

	// Get result
	ASSERT( !p.IsRunning() );
	int result = p.WaitForExit();

	if ( memErr.Get() )
	{ 
		DumpOutput( memErr.Get(), memErrSize );
	}

	if ( result != 0 )
	{
		DumpOutput( memOut.Get(), memOutSize );
		FLOG_ERROR( "Failed to build Object (error %i) '%s'", result, GetName().Get() );

		// for remote jobs, we must serialize the errors to return with the job
		if ( job->IsLocal() == false )
		{
			AutoPtr< char > mem( (char *)::Alloc( memOutSize + memErrSize ) );
			memcpy( mem.Get(), memOut.Get(), memOutSize );
			memcpy( mem.Get() + memOutSize, memErr.Get(), memErrSize );
			job->OwnData( mem.Release(), ( memOutSize + memErrSize ) );
		}

		return false;
	}

	return true;
}

//------------------------------------------------------------------------------
